import paho.mqtt.client as mqtt
import time
import os
import json
from threading import Thread
from diskcache import Deque

import uuid

from .singleton_wrapper import Singleton


class MQTTClient(Singleton):
    def __init__(self, client_id, username, password, local_cache_path='../mqtt_offlines'):
        self.connected = False

        self.client = mqtt.Client(client_id=client_id + str(uuid.uuid1()), clean_session=False)
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message
        self.client.on_publish = self._on_publish
        self.client.username_pw_set(username, password)

        self.subscribes = []

        # FileIO.create_dir_if_not_exists(local_cache_path)
        os.makedirs(local_cache_path, exist_ok=True)
        self.offlines = Deque(directory=local_cache_path)

    def connect(self, host, port):
        self.thread = Thread(target=self.__mqtt_connect, args=(host, port))
        self.thread.setDaemon(True)
        self.thread.start()
        return self

    def subscribe(self, topic, qos):
        self.subscribes.append((topic, qos))

    def register_on_message_handler(self, handler):
        self.on_message_handler = handler

    def join(self):
        if self.thread:
            self.thread.join()

    def __mqtt_connect(self, host, port):
        try:
            self.client.connect(host, port=port, keepalive=60)
            self.client.loop_forever()
        except Exception as e:
            print(e)

    def _on_connect(self, client, userdata, flags, rc):
        if not self.connected:
            self.connected = True
            print("mqtt service connected")
            
            for subscribe in self.subscribes:
                print('subscribing topic: {}/qos:{}...'.format(subscribe[0], subscribe[1]))
                self.client.subscribe(subscribe[0], subscribe[1])

            while len(self.offlines) > 0:
                print('sending offlines...')
                o_data = self.offlines.pop()
                try:
                    if o_data['expires'] >= time.time() and o_data['retries'] < 5:
                        self.client.publish(o_data['topic'], o_data['message'], qos=0)
                except Exception as e:
                    o_data['retries'] += 1
                    self.offlines.append(o_data)

    def _on_disconnect(self, client, userdata, rc):
        self.connected = False
        print("mqtt service disconnected")

    def _on_message(self, client, userdata, msg):
        if self.on_message_handler:
            self.on_message_handler(client, userdata, msg)

    def _on_publish(self, client, userdata, msg):
        print('published')

    def publish(self, message, topic, retain=False, qos=0, expires=7200):
        if not isinstance(message, str):
            message = json.dumps(message)

        tries = 0
        while not self.connected and tries < 5:
            time.sleep(5)
            tries += 1
        
        if tries >= 5:
            self.offlines.append({
                'topic': topic,
                'message': message,
                'expires': time.time() + expires * 1000,
                'retries': 0
            })
        else:
            try:
                self.client.publish(topic, payload=message, retain=retain, qos=qos)
                return True
            except Exception as e:
                pass
            return False
            


try:
    from contextlib import AsyncExitStack
except:
    from async_exit_stack import AsyncExitStack
import traceback
from minio import Minio

from aiobotocore.session import get_session
from aiofile import async_open


class MinioAgent:
    def __init__(self, endpoint, acs_key, secret_key, secure=True, location='default'):
        self.client = Minio(endpoint, access_key=acs_key,
                            secret_key=secret_key, secure=secure, region=location)
        self.location = location

    def create_bucket_if_not_exists(self, bucket_name, location):
        try:
            if not self.client.bucket_exists(bucket_name):
                self.client.make_bucket(bucket_name, location=location)

            return True
        except Exception as e:
            traceback.print_exc()
            return False

    def upload_file(self, path, bucket, bucket_filename, tries=0):
        try:
            if self.create_bucket_if_not_exists(bucket, self.location):
                self.client.fput_object(bucket, bucket_filename, path)
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                print('創建bucket失敗: {}'.format(bucket))
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file(path, bucket, bucket_filename, tries+1)

    def upload_file_by_buffer(self, buffer, bucket, bucket_filename, tries=0):
        try:
            if self.create_bucket_if_not_exists(bucket, self.location):
                self.client.put_object(
                    bucket, bucket_filename, buffer, len(buffer.getvalue()))
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file_by_buffer(buffer, bucket, bucket_filename, tries+1)

    def download_file(self, path, bucket, bucket_filename, tries=0):
        try:
            self.client.fget_object(bucket, bucket_filename, path)
            return True, path
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.download_file(path, bucket, bucket_filename, tries+1)

    def delete(self, bucket, bucket_name, tries):
        try:
            self.client.remove_object(bucket, bucket_name)
            return True, None
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.delete(bucket, bucket_name, tries + 1)

    def exists(self, bucket, bucket_name):
        meta = self.client.stat_object(bucket, bucket_name)

        return meta is not None


class MinioAgentAsync:
    def __init__(self, endpoint, acs_key, secret_key, secure=True, location='default', loop=None):
        self._exit_stack = AsyncExitStack()
        
        print('EndPoint: {}'.format(endpoint))
        session = get_session()
        self.client = loop.run_until_complete(self._exit_stack.enter_async_context(
            session.create_client('s3',
                                  use_ssl=False,
                                  verify=False,
                                  aws_secret_access_key=secret_key,
                                  aws_access_key_id=acs_key,
                                  endpoint_url=endpoint)
        ))
        self.location = location

    async def create_bucket_if_not_exists(self, bucket_name, location):
        try:
            await self.client.create_bucket(Bucket=bucket_name)

            return True
        except Exception as e:
            print(e)
            if 'BucketAlreadyOwnedByYou' in str(type(e)):
                return True

            return False

    async def upload_file(self, path, bucket, bucket_filename, tries=0):
        try:
            if await self.create_bucket_if_not_exists(bucket, self.location):
                resp = await self.client.put_object(Bucket=bucket, Key=bucket_filename, Body=open(path, 'rb'))
                if 'ETag' in resp:
                    return True, 'minio://{}/{}'.format(bucket, bucket_filename)
                else:
                    return False, '上传文件错误'
            else:
                print('創建bucket失敗: {}'.format(bucket))
                return False, '创建bucket失败'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file(path, bucket, bucket_filename, tries+1)

    async def upload_file_by_buffer(self, buffer, bucket, bucket_filename, tries=0):
        try:
            if self.create_bucket_if_not_exists(bucket, self.location):
                resp = await self.client.put_object(Bucket=bucket, Key=bucket_filename, Body=buffer)
                if 'ETag' in resp:
                    return True, 'minio://{}/{}'.format(bucket, bucket_filename)
                else:
                    return False, '上传错误'
            else:
                return False, '创建bucket失败'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file_by_buffer(buffer, bucket, bucket_filename, tries+1)

    async def download_file(self, path, bucket, bucket_filename, tries=0):
        try:
            response = await self.client.get_object(Bucket=bucket, Key=bucket_filename)
            async with async_open(path, 'wb') as f:
                async with response['Body'] as stream:
                    await f.write(await stream.read())
            return True, path
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.download_file(path, bucket, bucket_filename, tries+1)

    async def delete(self, bucket, bucket_name, tries):
        try:
            await self.client.delete_object(Bucket=bucket, Key=bucket_name)
            return True, None
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.delete(bucket, bucket_name, tries + 1)

    async def exists(self, bucket, bucket_name):
        meta = await self.client.get_object_acl(bucket, bucket_name)

        return meta is not None



from .nsqmanager import MQManager
from .redisutil import XRedis
from .singleton_wrapper import Singleton
from .minio import MinioAgent
from .mqtt import MQTTClient
from .cachemanager import CacheManager


from sqlite3 import DataError
import traceback
import httpx
from aiohttp_retry import ExponentialRetry, RetryClient
from aiofile import async_open

from qiniu import put_file, put_data, Auth, etag
from qiniu import BucketManager, build_batch_delete
from qiniu import Zone, set_default

from hamunafs.backends.base import BackendBase

class Qiniu(BackendBase):
    def __init__(self, cfg):
        key, secret, domain, default_bucket = cfg['key'], cfg['secret'], cfg['domain'], cfg['default_bucket']
        self.auth = Auth(key, secret)
        self.domain = domain
        self.default_bucket = default_bucket
        self.bucket = BucketManager(self.auth)
    
    def get_token(self, filename):
        return self.auth.upload_token(self.default_bucket, filename)
    
    def geturl(self, entrypoint):
        bucket, bucket_name = entrypoint.split('/')
        return 'http://{}/{}_{}'.format(self.domain, bucket, bucket_name)

    def put(self, file, bucket, bucket_name, tmp=True):
        try:
            if tmp:
                _bucket = 'tmp_file_' + bucket
            else:
                _bucket = bucket
            b_name = '{}_{}'.format(_bucket, bucket_name)
            print('acquiring upload token....')
            token = self.auth.upload_token(self.default_bucket, b_name)
            print('upload token acquired -> {}'.format(token))
            print('uploading...')
            ret, info = put_file(token, b_name, file)
            if ret is not None:
                print('upload success.')
                return True, '{}/{}'.format(_bucket, bucket_name)
            print('upload failed.')
            return False, '上传失败'
        except Exception as e:
            traceback.print_exc()
            return False, str(e)
    
    async def put_async(self, file, bucket, bucket_name, tmp=True):
        return self.put(file, bucket, bucket_name, tmp)

    def put_buffer(self, buffer, bucket, bucket_name):
        try:
            b_name = '{}_{}'.format(bucket, bucket_name)
            token = self.auth.upload_token(self.default_bucket, b_name)
            ret, info = put_data(token, b_name, buffer)
            if ret is not None:
                return True, '{}/{}'.format(bucket, bucket_name)
            return False, '上传失败'
        except Exception as e:
            return False, str(e)

    async def put_buffer_async(self, buffer, bucket, bucket_name):
        return self.put_buffer(buffer, bucket, bucket_name)

    def get(self, download_path, bucket, bucket_name, tries=0):
        try:
            if tries >= 3:
                return False, '下载出错'
            else:
                url = 'http://{}/{}'.format(self.domain, '{}_{}'.format(bucket, bucket_name))
                print(url)
                with httpx.stream('GET', url) as response:
                    if response.status_code == 200:
                        with open(download_path, mode='wb') as f:
                            for chunk in response.iter_bytes():
                                f.write(chunk)
                    else:
                        raise DataError()
                return True, download_path
        except Exception as e:
            if tries >= 3:
                return False, str(e)
            else:
                return self.get(download_path, bucket, bucket_name, tries+1)

    async def get_async(self, download_path, bucket, bucket_name):
        try:
            url = 'http://{}/{}'.format(self.domain, '{}_{}'.format(bucket, bucket_name))
            print(url)
            retry_opts = ExponentialRetry(attempts=3)
            async with RetryClient(retry_options=retry_opts) as session:
                async with session.get(url) as response:
                    if response.status == 200:
                        async with async_open(download_path, mode='wb') as f:
                            while True:
                                chunk = await response.content.read(4096)
                                if not chunk:
                                    break
                                await f.write(chunk)
                    else:
                        raise DataError()
            return True, download_path
        except Exception as e:
            return False, str(e)


            



# from hamunafs.utils.redisutil import XRedisAsync

# import asyncio

# redis = XRedisAsync('cache.ai.hamuna.club', '1987yang', 6379)

# asyncio.run(redis.zadd('1', '2', 3))
import asyncio
from aiobotocore.session import get_session
from botocore.client import Config
from aiofile import async_open
from async_exit_stack import AsyncExitStack

async def main():
    session = get_session()
    async with session.create_client('s3',
                                use_ssl=False,
                                aws_secret_access_key='1987yang',
                                aws_access_key_id='hmcz', 
                                config=Config(signature_version="s3v4"),
                                endpoint_url='http://localhost:9000') as client:
        resp = await client.put_object(Bucket='tmps',
                                            Key='temp1.txt',
                                            Body='test'.encode())
        resp = await client.delete_object(Bucket='tmps', Key='temp1.txt')
        pass

loop = asyncio.get_event_loop()
loop.run_until_complete(main())
